import uuid
import unittest
import decimal
import pandas as pd
from komlogd.api.protocol.model import messages
from komlogd.api.protocol.model.types import Metrics, Actions, Metric, Datasource, Datapoint

class ApiProtocolModelMessagesTest(unittest.TestCase):

    def test_version(self):
        ''' Protocol version '''
        self.assertEqual(messages.KomlogMessage._version_,1)

    def test_KomlogMessage_failure_cannot_instantiate_directly(self):
        ''' Direct instantiation of KomlogMessage objects is not allowed '''
        with self.assertRaises(TypeError) as cm:
            msg=messages.KomlogMessage()

    def test_KomlogMessage_version_modification_not_allowed(self):
        ''' We cannot modify the version in a KomlogMessage derived class '''
        uri='uri'
        ts=pd.Timestamp('now',tz='utc')
        content='content'
        msg=messages.SendDsData(uri=uri, ts=ts, content=content)
        with self.assertRaises(TypeError) as cm:
            msg.v=2

    def test_KomlogMessage_action_modification_not_allowed(self):
        ''' We cannot modify the action in a KomlogMessage derived class '''
        uri='uri'
        ts=pd.Timestamp('now',tz='utc')
        content='content'
        msg=messages.SendDsData(uri=uri, ts=ts, content=content)
        with self.assertRaises(TypeError) as cm:
            msg.action=Actions.SEND_MULTI_DATA

    def test_KomlogMessage_load_from_dict_failure_message_not_supported_no_dict(self):
        ''' KomlogMessage.load_from_dict should fail if msg passed is not a dict '''
        msg='message'
        with self.assertRaises(TypeError) as cm:
            messages.KomlogMessage.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Message not supported')

    def test_KomlogMessage_load_from_dict_failure_message_not_supported_no_action(self):
        ''' KomlogMessage.load_from_dict should fail if msg passed has no action field '''
        msg={'payload':'payload'}
        with self.assertRaises(TypeError) as cm:
            messages.KomlogMessage.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Message not supported')

    def test_KomlogMessage_load_from_dict_failure_unknown_action(self):
        ''' KomlogMessage.load_from_dict should fail if msg action is not in catalog '''
        msg={'action':'my_custom_action','payload':'payload'}
        with self.assertRaises(TypeError) as cm:
            messages.KomlogMessage.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Unknown message type')

    def test_KomlogMessage_load_from_dict_failure_not_implemented_in_this_version(self):
        ''' KomlogMessage.load_from_dict should fail if msg type has not implemented load_from_dict'''
        for action in (Actions.SEND_DS_DATA, Actions.SEND_DP_DATA, Actions.HOOK_TO_URI, Actions.UNHOOK_FROM_URI):
            msg={'action':action.value,'payload':'payload'}
            with self.assertRaises(NotImplementedError) as cm:
                messages.KomlogMessage.load_from_dict(msg)

    def test_SendDsData_failure_invalid_uri(self):
        ''' SendDsData creation should fail if uri is not valid '''
        uri='..not_valid'
        ts=pd.Timestamp('now',tz='utc')
        content='content'
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDsData(uri=uri, ts=ts, content=content)

    def test_SendDsData_failure_invalid_ts(self):
        ''' SendDsData creation should fail if ts is not valid '''
        uri='valid.uri'
        ts=-1.232
        content='content'
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDsData(uri=uri, ts=ts, content=content)

    def test_SendDsData_failure_invalid_content(self):
        ''' SendDsData creation should fail if content is not valid '''
        uri='valid.uri'
        ts=pd.Timestamp('now',tz='utc')
        content=223
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDsData(uri=uri, ts=ts, content=content)

    def test_SendDsData_failure_invalid_sequence(self):
        ''' SendDsData creation should fail if sequence is not valid '''
        uri='valid.uri'
        ts=pd.Timestamp('now',tz='utc')
        content='223'
        seq=uuid.uuid1()
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDsData(uri=uri, ts=ts, content=content, seq=seq)
        self.assertEqual(str(cm.exception),'Invalid sequence') 

    def test_SendDsData_failure_invalid_irt(self):
        ''' SendDsData creation should fail if irt is not valid '''
        uri='valid.uri'
        ts=pd.Timestamp('now',tz='utc')
        content='223'
        irt=uuid.uuid1()
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDsData(uri=uri, ts=ts, content=content, irt=irt)
        self.assertEqual(str(cm.exception),'Invalid irt') 

    def test_SendDsData_success(self):
        ''' SendDsData creation should succeed if parameters are valid '''
        uri='valid.uri'
        ts=pd.Timestamp('now',tz='utc')
        content='content\n'
        msg=messages.SendDsData(uri=uri, ts=ts, content=content)
        self.assertTrue(isinstance(msg, messages.SendDsData))
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.action, Actions.SEND_DS_DATA)
        self.assertEqual(msg.uri,uri)
        self.assertEqual(msg.ts,ts)
        self.assertEqual(msg.content,content)
        self.assertEqual(msg.irt,None)
        self.assertEqual(len(msg.seq),20)

    def test_SendDsData_failure_cannot_modify_sequence(self):
        ''' an exceptions should be raised if we try to modify a SendDsData object sequence param'''
        uri='valid.uri'
        ts=pd.Timestamp('now',tz='utc')
        content='content\n'
        msg=messages.SendDsData(uri=uri, ts=ts, content=content)
        self.assertTrue(isinstance(msg, messages.SendDsData))
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.action, Actions.SEND_DS_DATA)
        self.assertEqual(msg.uri,uri)
        self.assertEqual(msg.ts,ts)
        self.assertEqual(msg.content,content)
        with self.assertRaises(TypeError) as cm:
            msg.seq=uuid.uuid1().hex[0:20]
        self.assertEqual(str(cm.exception), 'Sequence cannot be modified')

    def test_SendDsData_to_dict_success(self):
        ''' SendDsData should return a valid dict representation of the object '''
        uri='valid.uri'
        ts=pd.Timestamp('now',tz='utc')
        content='content\n'
        msg=messages.SendDsData(uri=uri, ts=ts, content=content)
        self.assertTrue(isinstance(msg, messages.SendDsData))
        self.assertEqual(
            msg.to_dict(),{
                'v':messages.KomlogMessage._version_,
                'action':Actions.SEND_DS_DATA.value,
                'seq':msg.seq,
                'irt':None,
                'payload':
                    {'uri':uri,'ts':ts.isoformat(),'content':content}
                }
        )

    def test_SendDpData_failure_invalid_uri(self):
        ''' SendDpData creation should fail if uri is not valid '''
        uri='..not_valid'
        ts=pd.Timestamp('now',tz='utc')
        content='content'
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDpData(uri=uri, ts=ts, content=content)

    def test_SendDpData_failure_invalid_ts(self):
        ''' SendDpData creation should fail if ts is not valid '''
        uri='valid.uri'
        ts=-1.232
        content='content'
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDpData(uri=uri, ts=ts, content=content)

    def test_SendDpData_failure_invalid_content(self):
        ''' SendDpData creation should fail if content is not valid '''
        uri='valid.uri'
        ts=pd.Timestamp('now',tz='utc')
        content='223,223'
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDpData(uri=uri, ts=ts, content=content)

    def test_SendDpData_failure_invalid_sequence(self):
        ''' SendDpData creation should fail if sequence is not valid '''
        uri='valid.uri'
        ts=pd.Timestamp('now',tz='utc')
        content=223.223
        seq=uuid.uuid1().hex[0:10]
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDpData(uri=uri, ts=ts, content=content, seq=seq)
        self.assertEqual(str(cm.exception), 'Invalid sequence')

    def test_SendDpData_failure_invalid_irt(self):
        ''' SendDpData creation should fail if irt is not valid '''
        uri='valid.uri'
        ts=pd.Timestamp('now',tz='utc')
        content=223.223
        irt=uuid.uuid1().hex[0:10]
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDpData(uri=uri, ts=ts, content=content, irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid irt')

    def test_SendDpData_success(self):
        ''' SendDpData creation should succeed if parameters are valid '''
        uri='valid.uri'
        ts=pd.Timestamp('now',tz='utc')
        content='123.32'
        msg=messages.SendDpData(uri=uri, ts=ts, content=content)
        self.assertTrue(isinstance(msg, messages.SendDpData))
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.action, Actions.SEND_DP_DATA)
        self.assertEqual(msg.uri, uri)
        self.assertEqual(msg.ts,ts)
        self.assertEqual(msg.content,decimal.Decimal(content))
        self.assertEqual(msg.irt,None)
        self.assertEqual(len(msg.seq),20)

    def test_SendDpData_to_dict_success(self):
        ''' SendDpData should return a valid dict representation of the object '''
        uri='valid.uri'
        ts=pd.Timestamp('now',tz='utc')
        content=decimal.Decimal('300.2')
        msg=messages.SendDpData(uri=uri, ts=ts, content=content)
        self.assertTrue(isinstance(msg, messages.SendDpData))
        self.assertEqual(
            msg.to_dict(),{
                'v':messages.KomlogMessage._version_,
                'action':Actions.SEND_DP_DATA.value,
                'seq':msg.seq,
                'irt':None,
                'payload':
                    {'uri':uri,'ts':ts.isoformat(),'content':str(content)}
                }
        )

    def test_SendMultiData_failure_none_uris(self):
        ''' SendMultiData creation should fail if uris is None '''
        uris=None
        ts=pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts)

    def test_SendMultiData_failure_invalid_uris_invalid_type(self):
        ''' SendMultiData creation should fail if uris is not a list '''
        uris='uri'
        ts=pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts)

    def test_SendMultiData_failure_invalid_uris_invalid_item_type(self):
        ''' SendMultiData creation should fail if uris item is not a dict '''
        uris=['uri']
        ts=pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts)

    def test_SendMultiData_failure_invalid_uris_uri_key_not_found(self):
        ''' SendMultiData creation should fail if uris item has not uri '''
        uris=[{'type':Metrics.DATASOURCE,'content':'content '}]
        ts=pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts)

    def test_SendMultiData_failure_invalid_uris_content_key_not_found(self):
        ''' SendMultiData creation should fail if uris item has not uri '''
        uris=[{'uri':'uri','type':Metrics.DATAPOINT}]
        ts=pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts)

    def test_SendMultiData_failure_invalid_uris_type_key_not_found(self):
        ''' SendMultiData creation should fail if uris item has not uri '''
        uris=[{'uri':'uri','content':'content'}]
        ts=pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts)

    def test_SendMultiData_failure_invalid_uris_invalid_item_uri(self):
        ''' SendMultiData creation should fail if uris item has invalid uri '''
        uris=[{'uri':'invalid uri','type':Metrics.DATASOURCE,'content':'content'}]
        ts=pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts)

    def test_SendMultiData_failure_invalid_uris_invalid_item_content_for_metric_datasource(self):
        ''' SendMultiData creation should fail if uris item has invalid uri '''
        uris=[{'uri':'valid_uri','type':Metrics.DATASOURCE, 'content':{'content'}}]
        ts=pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts)

    def test_SendMultiData_failure_invalid_uris_invalid_item_content_for_metric_datapoint(self):
        ''' SendMultiData creation should fail if uris item has invalid uri '''
        uris=[{'uri':'valid_uri','type':Metrics.DATAPOINT, 'content':'content'}]
        ts=pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts)

    def test_SendMultiData_failure_invalid_ts(self):
        ''' SendMultiData creation should fail if ts is not valid '''
        uris=[{'uri':'uri','content':'333','type':Metrics.DATAPOINT}]
        ts=-1.232
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts)

    def test_SendMultiData_failure_none_ts(self):
        ''' SendMultiData creation should fail if ts is None '''
        uris=[{'uri':'uri','content':'333','type':Metrics.DATAPOINT}]
        ts=None
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts)

    def test_SendMultiData_failure_invalid_seq(self):
        ''' SendMultiData creation should fail if ts is not valid '''
        uris=[{'uri':'uri','content':'333','type':Metrics.DATAPOINT}]
        ts=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:10]
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts, seq=seq)
        self.assertEqual(str(cm.exception), 'Invalid sequence')

    def test_SendMultiData_failure_invalid_irt(self):
        ''' SendMultiData creation should fail if ts is not valid '''
        uris=[{'uri':'uri','content':'333','type':Metrics.DATAPOINT}]
        ts=pd.Timestamp('now',tz='utc')
        irt=uuid.uuid1().hex[0:10]
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData(uris=uris, ts=ts, irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid irt')

    def test_SendMultiData_success(self):
        ''' SendMultiData creation should succeed if parameters are valid '''
        uris=[
            {'uri':'uri.dp','content':decimal.Decimal('333'),'type':Metrics.DATAPOINT},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        message_uris=[
            {'uri':'uri.dp','content':decimal.Decimal('333'),'type':Metrics.DATAPOINT},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE},
        ]
        ts=pd.Timestamp('now',tz='utc')
        msg=messages.SendMultiData(uris=uris, ts=ts)
        self.assertTrue(isinstance(msg, messages.SendMultiData))
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.action, Actions.SEND_MULTI_DATA)
        self.assertEqual(sorted(msg.uris, key=lambda x: x['uri']), sorted(message_uris, key=lambda x: x['uri']))
        self.assertEqual(msg.ts,ts)
        self.assertEqual(msg.irt,None)
        self.assertEqual(len(msg.seq),20)

    def test_SendMultiData_success_extra_uris_fields_not_propagated(self):
        ''' SendMultiData creation should succeed and non standard uris keys should not be propagated '''
        uris=[
            {'uri':'uri.dp','content':'333','type':Metrics.DATAPOINT,'non_standard':'content'},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE,'non_standard':'content'},
        ]
        standard_uris=[
            {'uri':'uri.dp','content':decimal.Decimal('333'),'type':Metrics.DATAPOINT},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE},
        ]
        ts=pd.Timestamp('now',tz='utc')
        msg=messages.SendMultiData(uris=uris, ts=ts)
        self.assertTrue(isinstance(msg, messages.SendMultiData))
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.action, Actions.SEND_MULTI_DATA)
        self.assertEqual(sorted(msg.uris, key=lambda x:x['uri']), sorted(standard_uris, key=lambda x: x['uri']))
        self.assertEqual(msg.ts,ts)
        self.assertEqual(msg.irt,None)
        self.assertEqual(len(msg.seq),20)

    def test_SendMultiData_to_dict_success(self):
        ''' SendMultiData should return a valid dict representation of the object '''
        self.maxDiff=None
        uris=[
            {'uri':'uri.dp','content':decimal.Decimal('0.333'),'type':Metrics.DATAPOINT},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE},
        ]
        dict_uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        ts=pd.Timestamp('now',tz='utc')
        msg=messages.SendMultiData(uris=uris, ts=ts)
        self.assertTrue(isinstance(msg, messages.SendMultiData))
        dict_msg=msg.to_dict()
        dict_msg['payload']['uris']=sorted(dict_msg['payload']['uris'], key=lambda x:x['uri'])
        self.assertEqual(
            dict_msg,{
                'v':messages.KomlogMessage._version_,
                'action':Actions.SEND_MULTI_DATA.value,
                'seq':msg.seq,
                'irt':None,
                'payload':
                    {'uris':dict_uris,'ts':ts.isoformat()}
                }
        )

    def test_SendMultiData_load_from_dict_failure_no_dict_passed(self):
        ''' SendMultiData.load_from_dict should fail if parameter is not a dict '''
        data='data'
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendMultiData_load_from_dict_failure_no_version(self):
        ''' SendMultiData.load_from_dict should fail if data has no version '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'action':Actions.SEND_MULTI_DATA.value,'payload':
            {'uris':uris,'ts':ts.isoformat()},
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendMultiData_load_from_dict_failure_no_action(self):
        ''' SendMultiData.load_from_dict should fail if data has no action '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'payload':
            {'uris':uris,'ts':ts.isoformat()},
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendMultiData_load_from_dict_failure_no_seq(self):
        ''' SendMultiData.load_from_dict should fail if data has no seq '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_MULTI_DATA.value,
            'payload': {'uris':uris,'ts':ts.isoformat()},
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendMultiData_load_from_dict_failure_no_irt(self):
        ''' SendMultiData.load_from_dict should fail if data has no irt '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_MULTI_DATA.value,
            'payload': {'uris':uris,'ts':ts.isoformat()},
            'seq':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendMultiData_load_from_dict_failure_no_payload(self):
        ''' SendMultiData.load_from_dict should fail if data has no payload '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_MULTI_DATA.value,
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendMultiData_load_from_dict_failure_invalid_version(self):
        ''' SendMultiData.load_from_dict should fail if data version is not valid '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':'a', 'action':Actions.SEND_MULTI_DATA.value,'payload':
            {'uris':uris,'ts':ts.isoformat()},
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendMultiData_load_from_dict_failure_invalid_action(self):
        ''' SendMultiData.load_from_dict should fail if data has invalid action '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_DS_DATA.value,'payload':
            {'uris':uris,'ts':ts.isoformat()},
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendMultiData_load_from_dict_failure_payload_no_dict(self):
        ''' SendMultiData.load_from_dict should fail if payload is not a dict '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_MULTI_DATA.value,'payload':
            [{'uris':uris,'ts':ts.isoformat()},],
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendMultiData_load_from_dict_failure_payload_has_no_ts(self):
        ''' SendMultiData.load_from_dict should fail if payload has no ts '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_MULTI_DATA.value,'payload':
            {'uris':uris},
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendMultiData_load_from_dict_failure_payload_has_no_uris(self):
        ''' SendMultiData.load_from_dict should fail if payload has no uris '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_MULTI_DATA.value,'payload':
            {'ts':ts.isoformat()},
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendMultiData_load_from_dict_failure_invalid_ts(self):
        ''' SendMultiData.load_from_dict should fail if ts is invalid '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_MULTI_DATA.value,'payload':
            {'uris':uris,'ts':'1'},
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'ts type is not valid')

    def test_SendMultiData_load_from_dict_failure_invalid_uris(self):
        ''' SendMultiData.load_from_dict should fail if uris are invalid '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_MULTI_DATA.value,'payload':
            {'uris':uris,'ts':ts.isoformat()},
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Uris parameter not valid')

    def test_SendMultiData_load_from_dict_failure_invalid_seq(self):
        ''' SendMultiData.load_from_dict should fail if seq is invalid '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_MULTI_DATA.value,'payload':
            {'uris':uris,'ts':ts.isoformat()},
            'seq':uuid.uuid1().hex[0:10],
            'irt':uuid.uuid1().hex[0:20],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Invalid sequence')

    def test_SendMultiData_load_from_dict_failure_invalid_irt(self):
        ''' SendMultiData.load_from_dict should fail if irt is invalid '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_MULTI_DATA.value,'payload':
            {'uris':uris,'ts':ts.isoformat()},
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:10],
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(str(cm.exception), 'Invalid irt')

    def test_SendMultiData_load_from_dict_success(self):
        ''' SendMultiData.load_from_dict should succeed '''
        ts=pd.Timestamp('now',tz='utc')
        uris=[
            {'uri':'uri.dp','content':'0.333','type':Metrics.DATAPOINT.value},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE.value},
        ]
        obj_uris=[
            {'uri':'uri.dp','content':decimal.Decimal('0.333'),'type':Metrics.DATAPOINT},
            {'uri':'uri.ds','content':'content','type':Metrics.DATASOURCE},
        ]
        data={'v':messages.KomlogMessage._version_, 'action':Actions.SEND_MULTI_DATA.value,'payload':
            {'uris':uris,'ts':ts.isoformat()},
            'seq':uuid.uuid1().hex[0:20],
            'irt':uuid.uuid1().hex[0:20],
        }
        msg=messages.SendMultiData.load_from_dict(data)
        self.assertEqual(sorted(msg.uris, key=lambda x: x['uri']), sorted(obj_uris, key=lambda x: x['uri']))
        self.assertEqual(msg.ts, ts)
        self.assertEqual(msg.seq, data['seq'])
        self.assertEqual(msg.irt, data['irt'])
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.action, Actions.SEND_MULTI_DATA)

    def test_HookToUri_failure_none_uri(self):
        ''' HookToUri creation should fail if uri is None '''
        uri=None
        with self.assertRaises(TypeError) as cm:
            msg=messages.HookToUri(uri=uri)

    def test_HookToUri_failure_invalid_uri(self):
        ''' HookToUri creation should fail if uri is not valid '''
        uri='invalid uri'
        with self.assertRaises(TypeError) as cm:
            msg=messages.HookToUri(uri=uri)

    def test_HookToUri_failure_invalid_seq(self):
        ''' HookToUri creation should fail if seq is not valid '''
        uri='valid.uri'
        seq=123123
        irt=None
        with self.assertRaises(TypeError) as cm:
            msg=messages.HookToUri(uri=uri, seq=seq, irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid sequence')

    def test_HookToUri_failure_invalid_seq(self):
        ''' HookToUri creation should fail if irt is not valid '''
        uri='valid.uri'
        seq=uuid.uuid1().hex[0:20]
        irt=1
        with self.assertRaises(TypeError) as cm:
            msg=messages.HookToUri(uri=uri, seq=seq, irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid irt')

    def test_HookToUri_success(self):
        ''' HookToUri creation should succeed if parameters are valid '''
        uri='uri'
        seq=uuid.uuid1().hex[0:20]
        irt=None
        msg=messages.HookToUri(uri=uri, seq=seq, irt=irt)
        self.assertTrue(isinstance(msg, messages.HookToUri))
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.action, Actions.HOOK_TO_URI)
        self.assertEqual(msg.uri, uri)
        self.assertEqual(msg.irt, irt)
        self.assertEqual(msg.seq, seq)

    def test_HookToUri_to_dict_success(self):
        ''' HookToUri should return a valid dict representation of the object '''
        uri='uri'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        msg=messages.HookToUri(uri=uri, seq=seq, irt=irt)
        self.assertTrue(isinstance(msg, messages.HookToUri))
        self.assertEqual(
            msg.to_dict(), {
                'v':messages.KomlogMessage._version_,
                'action':Actions.HOOK_TO_URI.value,
                'seq':seq,
                'irt':irt,
                'payload': {'uri':uri}
            }
        )

    def test_UnHookFromUri_failure_none_uri(self):
        ''' UnHookFromUri creation should fail if uri is None '''
        uri=None
        with self.assertRaises(TypeError) as cm:
            msg=messages.UnHookFromUri(uri=uri)

    def test_UnHookFromUri_failure_invalid_uri(self):
        ''' UnHookFromUri creation should fail if uri is not valid '''
        uri='invalid uri'
        with self.assertRaises(TypeError) as cm:
            msg=messages.UnHookFromUri(uri=uri)

    def test_UnHookFromUri_failure_invalid_seq(self):
        ''' UnHookFromUri creation should fail if seq is not valid '''
        uri='valid.uri'
        seq=3
        irt=None
        with self.assertRaises(TypeError) as cm:
            msg=messages.UnHookFromUri(uri=uri, seq=seq, irt=irt)
        self.assertEqual(str(cm.exception),'Invalid sequence')

    def test_UnHookFromUri_failure_invalid_irt(self):
        ''' UnHookFromUri creation should fail if irt is not valid '''
        uri='valid.uri'
        seq=uuid.uuid1().hex[0:20]
        irt=1
        with self.assertRaises(TypeError) as cm:
            msg=messages.UnHookFromUri(uri=uri, seq=seq, irt=irt)
        self.assertEqual(str(cm.exception),'Invalid irt')

    def test_UnHookFromUri_success(self):
        ''' UnHookFromUri creation should succeed if parameters are valid '''
        uri='uri'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        msg=messages.UnHookFromUri(uri=uri, seq=seq, irt=irt)
        self.assertTrue(isinstance(msg, messages.UnHookFromUri))
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.action, Actions.UNHOOK_FROM_URI)
        self.assertEqual(msg.uri, uri)
        self.assertEqual(msg.seq, seq)
        self.assertEqual(msg.irt, irt)

    def test_UnHookFromUri_to_dict_success(self):
        ''' UnHookFromUri should return a valid dict representation of the object '''
        uri='uri'
        seq=None
        irt=None
        msg=messages.UnHookFromUri(uri=uri, seq=seq, irt=irt)
        self.assertTrue(isinstance(msg, messages.UnHookFromUri))
        self.assertNotEqual(msg.seq,None)
        self.assertEqual(
            msg.to_dict(),{
                'v':messages.KomlogMessage._version_,
                'action':Actions.UNHOOK_FROM_URI.value,
                'seq':msg.seq,
                'irt':irt,
                'payload': {'uri':uri}
            }
        )

    def test_RequestData_failure_invalid_uri(self):
        ''' creating a new RequestData instance should fail if uri is not valid '''
        uri='invalid uri'
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            messages.RequestData(uri=uri, start=start, end=end)
        self.assertEqual(str(cm.exception), 'uri is not valid: '+uri)

    def test_RequestData_failure_invalid_start(self):
        ''' creating a new RequestData instance should fail if start is not valid '''
        uri='valid.uri'
        start = 123123
        end = pd.Timestamp('now',tz='utc')
        with self.assertRaises(TypeError) as cm:
            messages.RequestData(uri=uri, start=start, end=end)
        self.assertEqual(str(cm.exception), 'ts type is not valid')

    def test_RequestData_failure_invalid_end(self):
        ''' creating a new RequestData instance should fail if end is not valid '''
        uri='valid.uri'
        start = pd.Timestamp('now',tz='utc')
        end = 123123
        with self.assertRaises(TypeError) as cm:
            messages.RequestData(uri=uri, start=start, end=end)
        self.assertEqual(str(cm.exception), 'ts type is not valid')

    def test_RequestData_failure_invalid_seq(self):
        ''' creating a new RequestData instance should fail if seq is not valid '''
        uri='valid.uri'
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        seq = uuid.uuid1().hex
        irt = uuid.uuid1().hex[0:20]
        with self.assertRaises(TypeError) as cm:
            messages.RequestData(uri=uri, start=start, end=end, seq=seq, irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid sequence')

    def test_RequestData_failure_invalid_irt(self):
        ''' creating a new RequestData instance should fail if irt is not valid '''
        uri='valid.uri'
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        seq = uuid.uuid1().hex[0:20]
        irt = uuid.uuid1().hex
        with self.assertRaises(TypeError) as cm:
            messages.RequestData(uri=uri, start=start, end=end, seq=seq, irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid irt')

    def test_RequestData_success(self):
        ''' creating a new RequestData instance should succeed '''
        uri='valid.uri'
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        msg=messages.RequestData(uri=uri, start=start, end=end, seq=seq, irt=irt)
        self.assertEqual(msg.uri, uri)
        self.assertEqual(msg.start, start)
        self.assertEqual(msg.end, end)
        self.assertEqual(msg.seq, seq)
        self.assertEqual(msg.irt, irt)

    def test_RequestData_to_dict_success(self):
        ''' RequestData.to_dict() should return a valid representation of the instance '''
        uri='valid.uri'
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        msg=messages.RequestData(uri=uri, start=start, end=end, seq=seq, irt=irt)
        self.assertEqual(msg.uri, uri)
        self.assertEqual(msg.start, start)
        self.assertEqual(msg.end, end)
        d=msg.to_dict()
        expected = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.REQUEST_DATA.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'count':None,
            }
        }
        self.assertEqual(d,expected)

    def test_SendDataInterval_failure_invalid_metric(self):
        ''' creating a new RequestData instance should fail if metric is not valid.
            Metric should be of type Datasource or Datapoint. '''
        metric=Metric(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        data = []
        with self.assertRaises(TypeError) as cm:
            messages.SendDataInterval(metric=metric, start=start, end=end, data=data)
        self.assertEqual(str(cm.exception), 'Invalid metric type')

    def test_SendDataInterval_failure_invalid_start(self):
        ''' creating a new RequestData instance should fail if start is not valid. '''
        metric=Datasource(uri='valid.uri')
        start = 213123
        end = pd.Timestamp('now',tz='utc')
        data = []
        with self.assertRaises(TypeError) as cm:
            messages.SendDataInterval(metric=metric, start=start, end=end, data=data)
        self.assertEqual(str(cm.exception), 'ts type is not valid')

    def test_SendDataInterval_failure_invalid_end(self):
        ''' creating a new RequestData instance should fail if end is not valid. '''
        metric=Datasource(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = 213123
        data = []
        with self.assertRaises(TypeError) as cm:
            messages.SendDataInterval(metric=metric, start=start, end=end, data=data)
        self.assertEqual(str(cm.exception), 'ts type is not valid')

    def test_SendDataInterval_failure_invalid_data_not_a_list(self):
        ''' creating a new RequestData instance should fail if data is not a list. '''
        metric=Datasource(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        data = tuple()
        with self.assertRaises(TypeError) as cm:
            messages.SendDataInterval(metric=metric, start=start, end=end, data=data)
        self.assertEqual(str(cm.exception), 'Invalid data')

    def test_SendDataInterval_failure_invalid_data_item_not_a_list(self):
        ''' creating a new RequestData instance should fail if a data item is not a list.'''
        metric=Datasource(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        data =[{'set'}]
        with self.assertRaises(TypeError) as cm:
            messages.SendDataInterval(metric=metric, start=start, end=end, data=data)
        self.assertEqual(str(cm.exception), 'Invalid data')

    def test_SendDataInterval_failure_invalid_data_item_does_not_have_two_items(self):
        ''' creating a new RequestData instance should fail if a data item does not have two items'''
        metric=Datasource(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'23232'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 253232323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 253232323','extra!'],
        ]
        with self.assertRaises(TypeError) as cm:
            messages.SendDataInterval(metric=metric, start=start, end=end, data=data)
        self.assertEqual(str(cm.exception), 'Invalid data')

    def test_SendDataInterval_failure_invalid_data_item_ts_is_invalid(self):
        ''' creating a new RequestData instance should fail if a data item ts is invalid '''
        metric=Datasource(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        data =[
            [234234,'23232'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 253232323'],
        ]
        with self.assertRaises(TypeError) as cm:
            messages.SendDataInterval(metric=metric, start=start, end=end, data=data)
        self.assertEqual(str(cm.exception), 'ts type is not valid')

    def test_SendDataInterval_failure_invalid_data_item_content_is_invalid_dp_content(self):
        ''' creating a new RequestData instance should fail if a data item content is invalid '''
        metric=Datapoint(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 253232323'],
        ]
        with self.assertRaises(TypeError) as cm:
            messages.SendDataInterval(metric=metric, start=start, end=end, data=data)
        self.assertEqual(str(cm.exception), 'datapoint value not valid')

    def test_SendDataInterval_failure_invalid_data_item_content_is_invalid_ds_content(self):
        ''' creating a new RequestData instance should fail if a data item content is invalid '''
        metric=Datasource(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),['ds content 323']],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 253232323'],
        ]
        with self.assertRaises(TypeError) as cm:
            messages.SendDataInterval(metric=metric, start=start, end=end, data=data)
        self.assertEqual(str(cm.exception), 'content is not a string')

    def test_SendDataInterval_failure_invalid_seq(self):
        ''' creating a new RequestData instance should fail if set is invalid '''
        metric=Datasource(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        seq = uuid.uuid1().hex
        irt = uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 253232323'],
        ]
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval(metric=metric, start=start, end=end, data=data, seq=seq, irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid sequence')

    def test_SendDataInterval_failure_invalid_irt(self):
        ''' creating a new RequestData instance should fail if set is invalid '''
        metric=Datasource(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        seq = uuid.uuid1().hex[0:20]
        irt = uuid.uuid1().hex[0:10]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 253232323'],
        ]
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval(metric=metric, start=start, end=end, data=data, seq=seq, irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid irt')

    def test_SendDataInterval_success_datasource(self):
        ''' creating a new RequestData instance should succeed '''
        metric=Datasource(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        seq = uuid.uuid1().hex[0:20]
        irt= uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'ds content 253232323'],
        ]
        msg=messages.SendDataInterval(metric=metric, start=start, end=end, data=data, seq=seq, irt=irt)
        self.assertEqual(msg.metric, metric)
        self.assertEqual(msg.start, start)
        self.assertEqual(msg.end, end)
        self.assertEqual(msg.seq, seq)
        self.assertEqual(msg.irt, irt)
        self.assertEqual(sorted(msg.data, key=lambda x: x[0]),sorted([(pd.Timestamp(item[0]),item[1]) for item in data], key=lambda x:x[0]))

    def test_SendDataInterval_success_datapoint(self):
        ''' creating a new RequestData instance should succeed '''
        metric=Datapoint(uri='valid.uri')
        start = pd.Timestamp('now',tz='utc')
        end = pd.Timestamp('now',tz='utc')
        seq = uuid.uuid1().hex[0:20]
        irt= uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'253232323'],
        ]
        msg=messages.SendDataInterval(metric=metric, start=start, end=end, data=data, seq=seq, irt=irt)
        self.assertEqual(msg.metric, metric)
        self.assertEqual(msg.start, start)
        self.assertEqual(msg.end, end)
        self.assertEqual(msg.seq, seq)
        self.assertEqual(msg.irt, irt)
        self.assertEqual(sorted(msg.data, key=lambda x: x[0]),sorted([(pd.Timestamp(item[0]),decimal.Decimal(item[1])) for item in data], key=lambda x:x[0]))

    def test_SendDataInterval_load_from_dict_failure_invalid_msg_type(self):
        ''' creating a new RequestData from a serialization should fail if serialization is not of type dict '''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = [{
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }]
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_no_version(self):
        ''' creating a new RequestData from a serialization should fail if msg has no version '''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'av':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_version(self):
        ''' creating a new RequestData from a serialization should fail if msg has invalid version '''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':[messages.KomlogMessage._version_],
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_no_action(self):
        ''' creating a new RequestData from a serialization should fail if msg has no action'''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'the_action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_action(self):
        ''' creating a new RequestData from a serialization should fail if msg has invalid action'''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.REQUEST_DATA.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_no_seq(self):
        ''' creating a new RequestData from a serialization should fail if msg has no seq'''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'the_seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_seq(self):
        ''' creating a new RequestData from a serialization should fail if msg has invalid seq'''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:10]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Invalid sequence')

    def test_SendDataInterval_load_from_dict_failure_no_irt(self):
        ''' creating a new RequestData from a serialization should fail if msg has no irt'''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'th_irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_irt(self):
        ''' creating a new RequestData from a serialization should fail if msg has invalid irt'''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:10]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Invalid irt')

    def test_SendDataInterval_load_from_dict_failure_no_payload(self):
        ''' creating a new RequestData from a serialization should fail if msg has no payload'''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'the_payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_payload_type(self):
        ''' creating a new RequestData from a serialization should fail if msg has invalid payload type'''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':[{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }]
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_payload_no_uri(self):
        ''' creating a new RequestData from a serialization should fail if msg payload has no uri '''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'ari':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_payload_uri_type(self):
        ''' creating a new RequestData from a serialization should fail if msg payload uri has invalid type '''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':[uri],
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_payload_uri_no_uri(self):
        ''' creating a new RequestData from a serialization should fail if msg payload uri has no uri '''
        uri={'ari':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_payload_uri_no_type(self):
        ''' creating a new RequestData from a serialization should fail if msg payload uri has no type '''
        uri={'uri':'valid.uri','taip':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_payload_no_start(self):
        ''' creating a new RequestData from a serialization should fail if payload has no start '''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'estart':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_payload_no_end(self):
        ''' creating a new RequestData from a serialization should fail if payload has no end '''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'theend':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_payload_no_data(self):
        ''' creating a new RequestData from a serialization should fail if payload has no data '''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'thedata':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Could not load message, invalid type')

    def test_SendDataInterval_load_from_dict_failure_invalid_payload_uri_type(self):
        ''' creating a new RequestData from a serialization should fail if payload has no data '''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        msg = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.SendDataInterval.load_from_dict(msg)
        self.assertEqual(str(cm.exception), 'Invalid metric type')

    def test_SendDataInterval_load_from_dict_success_datasource(self):
        ''' creating a new RequestData from a serialization should succeed '''
        uri={'uri':'valid.uri','type':Metrics.DATASOURCE.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'content 323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'content 253232323'],
        ]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        msg=messages.SendDataInterval.load_from_dict(source)
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.action, Actions.SEND_DATA_INTERVAL)
        self.assertEqual(msg.metric, Datasource(uri=uri['uri']))
        self.assertEqual(msg.start, pd.Timestamp(start))
        self.assertEqual(msg.end, pd.Timestamp(end))
        self.assertEqual(sorted(msg.data, key=lambda x:x[0]), sorted([(pd.Timestamp(item[0]),item[1]) for item in data]))

    def test_SendDataInterval_load_from_dict_success_datapoint(self):
        ''' creating a new SendDataInterval from a serialization should succeed '''
        uri={'uri':'valid.uri','type':Metrics.DATAPOINT.value}
        start=pd.Timestamp('now',tz='utc')
        end=pd.Timestamp('now',tz='utc')
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        data =[
            [pd.Timestamp('now',tz='utc').isoformat(),'323'],
            [pd.Timestamp('now',tz='utc').isoformat(),'3223'],
            [pd.Timestamp('now',tz='utc').isoformat(),'25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'25623'],
            [pd.Timestamp('now',tz='utc').isoformat(),'253232323'],
        ]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DATA_INTERVAL.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'uri':uri,
                'start':start.isoformat(),
                'end':end.isoformat(),
                'data':data
            }
        }
        msg=messages.SendDataInterval.load_from_dict(source)
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.action, Actions.SEND_DATA_INTERVAL)
        self.assertEqual(msg.metric, Datapoint(uri=uri['uri']))
        self.assertEqual(msg.start, pd.Timestamp(start))
        self.assertEqual(msg.end, pd.Timestamp(end))
        self.assertEqual(sorted(msg.data, key=lambda x:x[0]), sorted([(pd.Timestamp(item[0]),decimal.Decimal(item[1])) for item in data]))

    def test_GenericResponse_failure_invalid_status(self):
        ''' creating a GenericResponse obj should fail if status is invalid '''
        status='status'
        error=0
        reason=None
        seq=None
        irt=None
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse(status=status,error=error,reason=reason,seq=seq,irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid status')

    def test_GenericResponse_failure_invalid_error(self):
        ''' creating a GenericResponse obj should fail if error is invalid '''
        status=4200
        error=-1
        reason=None
        seq=None
        irt=None
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse(status=status,error=error,reason=reason,seq=seq,irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid error')

    def test_GenericResponse_failure_invalid_reason(self):
        ''' creating a GenericResponse obj should fail if reason is invalid '''
        status=4200
        error=0
        reason=2323
        seq=None
        irt=None
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse(status=status,error=error,reason=reason,seq=seq,irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid reason')

    def test_GenericResponse_failure_invalid_seq(self):
        ''' creating a GenericResponse obj should fail if seq is invalid '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex
        irt=None
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse(status=status,error=error,reason=reason,seq=seq,irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid sequence')

    def test_GenericResponse_failure_invalid_irt(self):
        ''' creating a GenericResponse obj should fail if irt is invalid '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse(status=status,error=error,reason=reason,seq=seq,irt=irt)
        self.assertEqual(str(cm.exception), 'Invalid irt')

    def test_GenericResponse_success(self):
        ''' creating a GenericResponse obj should succeed '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        msg=messages.GenericResponse(status=status,error=error,reason=reason,seq=seq,irt=irt)
        self.assertEqual(msg.action, Actions.GENERIC_RESPONSE)
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.status, status)
        self.assertEqual(msg.error, error)
        self.assertEqual(msg.reason, reason)
        self.assertEqual(msg.seq, seq)
        self.assertEqual(msg.irt, irt)

    def test_GenericResponse_load_from_dict_failure_no_version(self):
        ''' creating a new GenericResponse from a serialization should fail if it has no version'''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'the_v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Could not load message, invalid type')

    def test_GenericResponse_load_from_dict_failure_invalid_version(self):
        ''' creating a new GenericResponse from a serialization should fail if it version is invalid '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':'invalid',
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Could not load message, invalid type')

    def test_GenericResponse_load_from_dict_failure_no_action(self):
        ''' creating a new GenericResponse from a serialization should fail if it has no action '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'the_action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Could not load message, invalid type')

    def test_GenericResponse_load_from_dict_failure_invalid_action(self):
        ''' creating a new GenericResponse from a serialization should fail if it has invalid action '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.SEND_DS_DATA.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Could not load message, invalid type')

    def test_GenericResponse_load_from_dict_failure_no_seq(self):
        ''' creating a new GenericResponse from a serialization should fail if it has no seq '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'iseq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Could not load message, invalid type')

    def test_GenericResponse_load_from_dict_failure_invalid_seq(self):
        ''' creating a new GenericResponse from a serialization should fail if it has invalid seq '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:10]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Invalid sequence')

    def test_GenericResponse_load_from_dict_failure_no_irt(self):
        ''' creating a new GenericResponse from a serialization should fail if it has no irt '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'iirt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Could not load message, invalid type')

    def test_GenericResponse_load_from_dict_failure_invalid_irt(self):
        ''' creating a new GenericResponse from a serialization should fail if it has invalid irt '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:30]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Invalid irt')

    def test_GenericResponse_load_from_dict_failure_no_payload(self):
        ''' creating a new GenericResponse from a serialization should fail if it has no payload  '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'the_payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Could not load message, invalid type')

    def test_GenericResponse_load_from_dict_failure_invalid_payload(self):
        ''' creating a new GenericResponse from a serialization should fail if it has invalid  payload  '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':[{
                'status':status,
                'error':error,
                'reason':reason,
            }]
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Could not load message, invalid type')

    def test_GenericResponse_load_from_dict_failure_no_status(self):
        ''' creating a new GenericResponse from a serialization should fail if it has no status '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'the_status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Could not load message, invalid type')

    def test_GenericResponse_load_from_dict_failure_invalid_status(self):
        ''' creating a new GenericResponse from a serialization should fail if it has invalid status '''
        status='4200'
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Invalid status')

    def test_GenericResponse_load_from_dict_failure_no_error(self):
        ''' creating a new GenericResponse from a serialization should fail if it has no error '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'the_error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Could not load message, invalid type')

    def test_GenericResponse_load_from_dict_failure_invalid_error(self):
        ''' creating a new GenericResponse from a serialization should fail if it has invalid error '''
        status=4200
        error='0'
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Invalid error')

    def test_GenericResponse_load_from_dict_failure_no_reason(self):
        ''' creating a new GenericResponse from a serialization should fail if it has no reason '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'the_reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Could not load message, invalid type')

    def test_GenericResponse_load_from_dict_failure_invalid_reason(self):
        ''' creating a new GenericResponse from a serialization should fail if it has invalid reason '''
        status=4200
        error=0
        reason={'reason':'invalid_reason'}
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        with self.assertRaises(TypeError) as cm:
            msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(str(cm.exception),'Invalid reason')

    def test_GenericResponse_load_from_dict_success(self):
        ''' creating a new GenericResponse from a serialization should succeed '''
        status=4200
        error=0
        reason='reason'
        seq=uuid.uuid1().hex[0:20]
        irt=uuid.uuid1().hex[0:20]
        source = {
            'v':messages.KomlogMessage._version_,
            'action':Actions.GENERIC_RESPONSE.value,
            'seq':seq,
            'irt':irt,
            'payload':{
                'status':status,
                'error':error,
                'reason':reason,
            }
        }
        msg=messages.GenericResponse.load_from_dict(source)
        self.assertEqual(msg.v, messages.KomlogMessage._version_)
        self.assertEqual(msg.action, Actions.GENERIC_RESPONSE)
        self.assertEqual(msg.seq, seq)
        self.assertEqual(msg.irt, irt)
        self.assertEqual(msg.status, status)
        self.assertEqual(msg.error, error)
        self.assertEqual(msg.reason, reason)




'''

Main application

'''

import asyncio
import os
import sys
import signal
import functools
import traceback
from stat import S_ISFIFO, S_ISREG
from komlogd.base import config, logging
from komlogd.transfer_methods import main as tfmain
from komlogd.web import main as webmain

app = None

async def _signal_handler(signame):
    logging.logger.info(signame+' detected.')
    if app:
        await app.stop()

class Application:

    def __init__(self, config_file, uri):
        self.config_file = config_file
        self.uri = uri
        self._input_detected = True if S_ISFIFO(os.fstat(0).st_mode) or S_ISREG(os.fstat(0).st_mode) else False
        self._stdin_mode = True if uri else False
        try:
            config.load_application_config(filename=self.config_file)
            logging.initialize_logger()
            if self._input_detected != self._stdin_mode:
                if self._stdin_mode:
                    raise RuntimeError('uri parameter found, but no input detected')
                else:
                    raise RuntimeError('stdin data detected, but no uri parameter found')
            if self._stdin_mode:
                webmain.initialize_komlog_session(load_tm=False)
            else:
                tfmain.load_transfer_methods_files()
                webmain.initialize_komlog_session()
        except Exception as e:
            sys.stderr.write('Error initializing komlogd.\n')
            if logging.logger is not None and len(logging.logger.handlers)>0:
                sys.stderr.write('See log file: '+logging.logger.handlers[0].baseFilename+'\n')
                ex_info=traceback.format_exc().splitlines()
                for line in ex_info:
                    logging.logger.error(line)
            else:
                ex_info=traceback.format_exc().splitlines()
                for line in ex_info:
                    print (line)
            exit()

    async def start(self):
        logging.logger.debug('Starting komlogd.')
        if self._stdin_mode:
            await webmain.start_komlogd_stdin_mode(uri=self.uri)
        else:
            await webmain.start_komlog_session()

    async def stop(self):
        logging.logger.debug('Stopping komlogd.')
        await webmain.stop_komlog_session()

if sys.platform == 'win32':
    loop = asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)
else:
    loop = asyncio.get_event_loop()
    for signame in ('SIGINT', 'SIGTERM'):
        loop.add_signal_handler(getattr(signal, signame), lambda: asyncio.ensure_future(_signal_handler(signame)))

def start_application(config_file=None, uri=None):
    global app
    app = Application(config_file=config_file, uri=uri)
    try:
        loop.run_until_complete(app.start())
    except Exception as e:
        sys.stderr.write('Error detected.\n')
        if logging.logger:
            sys.stderr.write('See log file: '+logging.logger.handlers[0].baseFilename+'\n')
            ex_info=traceback.format_exc().splitlines()
            for line in ex_info:
                logging.logger.error(line)
    finally:
        logging.logger.debug('Closing loop.')
        loop.close()




import pandas as pd
from copy import deepcopy
from enum import Enum, unique
from komlogd.api.protocol.model import validation

@unique
class Actions(Enum):
    HOOK_TO_URI             = 'hook_to_uri'
    SEND_MULTI_DATA         = 'send_multi_data'
    SEND_DP_DATA            = 'send_dp_data'
    SEND_DS_DATA            = 'send_ds_data'
    UNHOOK_FROM_URI         = 'unhook_from_uri'
    REQUEST_DATA            = 'request_data'
    SEND_DATA_INTERVAL      = 'send_data_interval'
    GENERIC_RESPONSE        = 'generic_response'

@unique
class Metrics(Enum):
    DATASOURCE              = 'd'
    DATAPOINT               = 'p'

class Metric:
    _m_type_ = None

    def __init__(self, uri):
        self.uri = uri

    def __eq__(self, other):
        return self.uri == other.uri

    def __hash__(self):
        return hash(self.uri)

    def __copy__(self, memo):
        return type(self)(self.uri)

    def __deepcopy__(self, memo):
        return type(self)(deepcopy(self.uri, memo))

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, value):
        if not getattr(self, 'uri', None):
            validation.validate_uri(value)
            self._uri = value
            return
        raise TypeError('uri cannot be modified')

    @property
    def m_type(self):
        return self._m_type_

    @m_type.setter
    def m_type(self, value):
        raise TypeError('m_type cannot be modified')


class Datasource(Metric):
    _m_type_ = Metrics.DATASOURCE

    def __init__(self, uri):
        super().__init__(uri=uri)

class Datapoint(Metric):
    _m_type_ = Metrics.DATAPOINT

    def __init__(self, uri):
        super().__init__(uri=uri)

class Anomaly(Datapoint):

    def __init__(self, metric):
        super().__init__(uri='.'.join((metric.uri,'_anomaly')))

    def __deepcopy__(self, memo):
        uri = self.uri.split('._anomaly')[0]
        return type(self)(Metric(deepcopy(uri, memo)))

class Tag(Datapoint):

    def __init__(self, metric, key, value):
        super().__init__(uri='.'.join((metric.uri,'_tags',key,value)))

    def __deepcopy__(self, memo):
        uri,rest = self.uri.split('._tags.')[0:2]
        key,value = rest.split('.')[0:2]
        return type(self)(metric=Metric(deepcopy(uri, memo)),key=key, value=value)

class Sample:
    def __init__(self, metric, data, ts):
        self.metric = metric
        self.ts = ts
        self.data = data

    @property
    def metric(self):
        return self._metric

    @metric.setter
    def metric(self, value):
        if isinstance(value, Metric):
            self._metric = value
        else:
            raise TypeError('Invalid metric parameter')

    @property
    def ts(self):
        return self._ts

    @ts.setter
    def ts(self, ts):
        validation.validate_ts(ts)
        self._ts=pd.Timestamp(ts)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        if self.metric and self.metric.m_type == Metrics.DATASOURCE:
            validation.validate_ds_content(data)
        elif self.metric and self.metric.m_type == Metrics.DATAPOINT:
            validation.validate_dp_content(data)
        else:
            raise TypeError('Invalid Metric')
        self._data=data





'''
Templates file

'''

TEMPLATE_CONFIG_FILE='''
# komlogd configuration file

# username to connect to komlog. Set it and uncomment the following line:
#- username: username

# if you want to specify a different private key file, uncomment this line and set the filename:
#- keyfile: /home/komlogd/.komlogd/key.priv

# logging configuration, default values are indicated:
#- logging:
#    level: INFO
#    rotation: yes
#    max_bytes: 10000000
#    backup_count: 3
#    dirname: log
#    filename: komlogd.log


# Create an entry like the next one for every transfers file you want to load:
#- transfers:
#   enabled: yes
#   filename: /home/komlog/.komlogd/transfers/transfer_methods.py

# In the previous example, you can set the filename with the relative path from the application directory, too.
# Application directory is the one where the configuration file is located.
# So, if application directory is /home/komlog/.komlogd, we can set the previous filename as:
# filename: transfers/transfer_methods.py

'''




from enum import Enum, unique

@unique
class Status(int, Enum):
    # status regarding protocol

    PROTOCOL_ERROR                  = 4400
    ACCESS_DENIED                   = 4403

    # status regarding message

    MESSAGE_EXECUTION_OK            = 4200
    MESSAGE_ACCEPTED_FOR_PROCESSING = 4202
    MESSAGE_EXECUTION_DENIED        = 4430
    MESSAGE_EXECUTION_ERROR         = 4501

    # status regarding resources

    RESOURCE_NOT_FOUND              = 4404

    # status regarding server

    INTERNAL_ERROR                  = 4500
    SERVICE_UNAVAILABLE             = 4503




import asyncio
import json
import time
import pandas as pd
from komlogd.api import logging
from komlogd.api.protocol.model import messages
from komlogd.api.protocol.model.schedules import OnUpdateSchedule, CronSchedule
from komlogd.api.protocol.model.codes import Status
from komlogd.api.protocol.model.types import Metrics, Actions, Datasource, Datapoint, Metric, Sample

async def initialize_transfer_methods(session):
    methods = session._transfer_methods.get_transfer_methods(enabled=False)
    for method in methods:
        await initialize_transfer_method(session, method)

async def initialize_transfer_method(session, method):
    metrics = method.m_in
    if isinstance(method.schedule, OnUpdateSchedule):
        for m in method.schedule.metrics:
            metrics.add(m)
    initialized={metric.uri:False for metric in metrics}
    for metric in metrics:
        reqs = method.data_reqs[metric.uri] if isinstance(method.data_reqs, dict) and metric.uri in method.data_reqs else method.data_reqs
        if reqs:
            session._metrics_store.add_metric_data_reqs(metric=metric, reqs=reqs)
            end = pd.Timestamp('now',tz='utc') if reqs.past_delta else None
            start= end - reqs.past_delta if reqs.past_delta else None
            count = reqs.past_count if reqs.past_count else None
        msg = messages.HookToUri(uri=metric.uri)
        rsp = await session._await_response(msg)
        session._mark_message_done(msg.seq)
        if rsp.action == Actions.GENERIC_RESPONSE and rsp.status == Status.MESSAGE_EXECUTION_OK:
            if reqs is None:
                initialized[metric.uri]=True
            else:
                msg2 = messages.RequestData(uri=metric.uri, start=start, end=end, count=count)
                rsp2 = await session._await_response(msg2)
                done = False
                done_start = False if start else True
                done_end = False if end else True
                while not done:
                    if rsp2.action == Actions.GENERIC_RESPONSE:
                        if rsp2.status != Status.MESSAGE_ACCEPTED_FOR_PROCESSING:
                            session._mark_message_done(msg2.seq)
                            done = True
                            logging.logger.debug('Error requesting data for {}. {}'.format(str(metric.uri),str(rsp2.__dict__)))
                        else:
                            future = session._mark_message_undone(msg2.seq)
                            rsp2 = await future
                    elif rsp2.action == Actions.SEND_DATA_INTERVAL:
                        if start and rsp2.start == start:
                            done_start = True
                        if end and rsp2.end == end:
                            done_end = True
                        for row in rsp2.data[::-1]:
                            session._metrics_store.store(metric=rsp2.metric, ts=row[0], content=row[1])
                        if done_start and done_end:
                            session._mark_message_done(msg2.seq)
                            done = True
                            initialized[metric.uri]=True
                        else:
                            future = session._mark_message_undone(msg2.seq)
                            rsp2 = await future
        elif rsp.action == Actions.GENERIC_RESPONSE and rsp.status == Status.RESOURCE_NOT_FOUND:
            initialized[metric.uri]=True
        else:
            logging.logger.debug('Error hooking to uri {}. {}'.format(str(metric.uri),str(rsp.__dict__)))
    #by now, always initialize no matter if all metrics initialized ok. will add more parameters for fine grained initialization options
    if method.schedule.exec_on_load:
        now = pd.Timestamp('now', tz='utc')
        asyncio.ensure_future(exec_transfer_method(mid=method.mid, ts=now, metrics=[], session=session))
    if isinstance(method.schedule, CronSchedule):
        logging.logger.debug('Programming first transfer method exec: '+method.f.__name__)
        asyncio.ensure_future(session._periodic_transfer_method_call(mid=method.mid))
    if session._transfer_methods.enable_transfer_method(method.mid):
        logging.logger.debug('Transfer method initialized: {}'.format(str(method.f.__name__)))
    else:
        logging.logger.debug('Error initializing transfer method: {}'.format(str(method.f.__name__)))

async def send_samples(session, samples):
    if not isinstance(samples, list):
        return False
    grouped = {}
    for sample in samples:
        if isinstance(sample, Sample):
            try:
                grouped[sample.ts].append(sample)
            except KeyError:
                grouped[sample.ts]=[sample]
    logging.logger.debug('Grouped {}'.format(str((grouped))))
    tss = grouped.keys()
    msgs = []
    for ts in tss:
        items = grouped[ts]
        if len(items)>1:
            uris = []
            for item in items:
                uris.append({
                    'uri':item.metric.uri,
                    'type':item.metric.m_type.value,
                    'content':item.data,
                })
            msgs.append(messages.SendMultiData(ts=ts, uris=uris))
        elif isinstance(items[0].metric, Datasource):
            msgs.append(messages.SendDsData(uri=items[0].metric.uri, ts=ts, content=items[0].data))
        elif isinstance(items[0].metric, Datapoint):
            msgs.append(messages.SendDpData(uri=items[0].metric.uri, ts=ts, content=items[0].data))
        else:
            raise TypeError('invalid metric type')
    msgs.sort(key=lambda x: x.ts)
    result= {'success':True, 'error':''}
    for msg in msgs:
        rsp = await session._await_response(msg)
        session._mark_message_done(msg.seq)
        if rsp.status not in (Status.MESSAGE_ACCEPTED_FOR_PROCESSING, Status.MESSAGE_EXECUTION_OK):
            result['success']=False
            result['msg']=' '.join(('code:',rsp.error,rsp.reason))
    return result

async def exec_transfer_method(session, mid, ts, metrics):
    tm_info = session._transfer_methods.get_transfer_method_info(mid=mid)
    if tm_info:
        data = {}
        data_reqs = tm_info['tm'].get_data_requirements()
        for m,dr in data_reqs.items():
            its = None
            count = None
            if dr and dr.past_delta:
                its = ts - dr.past_delta
            if dr and not its and dr.past_count:
                count = dr.past_count
            data[m]=session._metrics_store.get_serie(metric=m, ets=ts, its=its, count=count)
        result = await tm_info['tm'].f(ts=ts, metrics=metrics, data=data)
        if result and 'samples' in result:
            samples = []
            for sample in result['samples']:
                if not session._metrics_store.isin(sample.metric, ts=sample.ts, content=sample.data):
                    samples.append(sample)
            await session.send_samples(samples)





import unittest
import decimal
import pandas as pd
import datetime
import uuid
from komlogd.api.protocol.model import types

class ApiProtocolModelTypesTest(unittest.TestCase):

    def test_actions(self):
        ''' Actions available in this protocol version '''
        self.assertEqual(len(types.Actions),8)
        self.assertEqual(types.Actions.SEND_DS_DATA.value,'send_ds_data')
        self.assertEqual(types.Actions.SEND_DP_DATA.value,'send_dp_data')
        self.assertEqual(types.Actions.SEND_MULTI_DATA.value,'send_multi_data')
        self.assertEqual(types.Actions.HOOK_TO_URI.value,'hook_to_uri')
        self.assertEqual(types.Actions.UNHOOK_FROM_URI.value,'unhook_from_uri')
        self.assertEqual(types.Actions.REQUEST_DATA.value,'request_data')
        self.assertEqual(types.Actions.SEND_DATA_INTERVAL.value,'send_data_interval')
        self.assertEqual(types.Actions.GENERIC_RESPONSE.value,'generic_response')

    def test_metrics(self):
        ''' Metrics available in this protocol version '''
        self.assertEqual(len(types.Metrics),2)
        self.assertEqual(types.Metrics.DATASOURCE.value,'d')
        self.assertEqual(types.Metrics.DATAPOINT.value,'p')

    def test_metric_failure_invalid_uri(self):
        ''' creating a Metric with invalid uri must fail '''
        uri='invalid uri'
        with self.assertRaises(TypeError) as cm:
            metric=types.Metric(uri=uri)
        self.assertEqual(str(cm.exception),'uri is not valid: '+uri)

    def test_metric_success(self):
        ''' creating a Metric object should succeed '''
        uri='valid.uri'
        metric=types.Metric(uri)
        self.assertTrue(isinstance(metric,types.Metric))
        self.assertEqual(metric.uri, uri)
        self.assertEqual(metric.m_type, None)

    def test_metric_uri_cannot_be_modified(self):
        ''' modifying a metric uri is not allowed '''
        uri='valid.uri'
        metric=types.Metric(uri)
        with self.assertRaises(TypeError) as cm:
            metric.uri='new.uri'
        self.assertEqual(str(cm.exception), 'uri cannot be modified')

    def test_metric_mtype_cannot_be_modified(self):
        ''' modifying a metric mtype is not allowed '''
        uri='valid.uri'
        metric=types.Metric(uri)
        with self.assertRaises(TypeError) as cm:
            metric.m_type=types.Metrics.DATASOURCE
        self.assertEqual(str(cm.exception), 'm_type cannot be modified')

    def test_datasource_metric_success(self):
        ''' creating a Datasource object should succeed '''
        uri='valid.uri'
        metric=types.Datasource(uri)
        self.assertTrue(isinstance(metric,types.Datasource))
        self.assertEqual(metric.uri, uri)
        self.assertEqual(metric.m_type, types.Metrics.DATASOURCE)

    def test_datapoint_metric_success(self):
        ''' creating a Datapoint object should succeed '''
        uri='valid.uri'
        metric=types.Datapoint(uri)
        self.assertTrue(isinstance(metric,types.Datapoint))
        self.assertEqual(metric.uri, uri)
        self.assertEqual(metric.m_type, types.Metrics.DATAPOINT)

    def test_comparing_two_metrics_is_done_by_uri(self):
        ''' when comparing two metrics, we check the uri field '''
        ds1=types.Metric('metrics.ds')
        ds2=types.Datasource('metrics.ds')
        self.assertEqual(ds1,ds2)

    def test_deleting_a_metric_from_an_array_is_done_checking_the_uri_field(self):
        ''' when comparing two metrics, we check the uri field '''
        metrics=[]
        metrics.append(types.Metric('metrics.uri1'))
        metrics.append(types.Metric('metrics.uri2'))
        metrics.append(types.Metric('metrics.uri3'))
        metrics.append(types.Metric('metrics.uri4'))
        metrics.append(types.Metric('metrics.uri5'))
        metrics.append(types.Metric('metrics.uri6'))
        self.assertEqual(len(metrics),6)
        existing_metric=types.Datapoint('metrics.uri1')
        non_existing_metric=types.Datapoint('metrics.uri10')
        self.assertTrue(existing_metric in metrics)
        self.assertFalse(non_existing_metric in metrics)
        metrics.remove(existing_metric)
        self.assertFalse(existing_metric in metrics)
        self.assertEqual(len(metrics),5)

    def test_metrics_are_hashable_objects(self):
        ''' a Metric is a hashable object, so you can use it as a key in dictionaries '''
        metrics={}
        for uri in range(1,10):
            metric=types.Metric(uri=str(uri))
            metrics[metric]=uri
        for uri in range(1,10):
            metric=types.Metric(uri=str(uri))
            self.assertEqual(metrics[metric],uri)




class ConfigLoadException(Exception):
    def __init__(self, msg=''):
        self.msg= msg

class LoggerException(Exception):
    def __init__(self, msg=''):
        self.msg= msg

class CryptographyException(Exception):
    def __init__(self, msg=''):
        self.msg= msg

class BadParametersException(Exception):
    def __init__(self, msg=''):
        self.msg= msg




'''

transfermethod decorator


'''

import asyncio
import copy
import inspect
import uuid
import pandas as pd
from functools import wraps
from komlogd.api import logging, uri, exceptions
from komlogd.api.model.transfer_methods import anon_transfer_methods
from komlogd.api.protocol.model import validation
from komlogd.api.protocol.model.schedules import Schedule, OnUpdateSchedule
from komlogd.api.protocol.model.types import Metric, Sample
from komlogd.api.protocol.model.transfer_methods import DataRequirements


class transfermethod:

    def __init__(self, p_in=None, p_out=None, data_reqs=None, schedule=None, allow_loops=False):
        self.mid = uuid.uuid4()
        self.last_exec = None
        self.p_in = p_in
        self.p_out = p_out
        self.data_reqs = data_reqs
        self.schedule = schedule
        self.allow_loops = allow_loops

    @property
    def p_in(self):
        return self._p_in

    @p_in.setter
    def p_in(self, value):
        if value is None:
            self._p_in = {}
        elif isinstance(value, dict):
            for k in value.keys():
                if k in ['ts','updated','others']:
                    raise exceptions.BadParametersException('Invalid input parameter. "{}" is a reserved parameter'.format(str(k)))
            else:
                self._p_in = value
        else:
            raise exceptions.BadParametersException('"p_in" attribute must be a dict')

    @property
    def p_out(self):
        return self._p_out

    @p_out.setter
    def p_out(self, value):
        if value is None:
            self._p_out = {}
        elif isinstance(value, dict):
            for k in value.keys():
                if k in ['ts','updated','others']:
                    raise exceptions.BadParametersException('Invalid output parameter. "{}" is a reserved parameter'.format(str(k)))
            else:
                self._p_out = value
        else:
            raise exceptions.BadParametersException('"p_out" attribute must be a dict')

    @property
    def schedule(self):
        return self._schedule

    @schedule.setter
    def schedule(self, value):
        if value is None or isinstance(value, Schedule):
            self._schedule = value
        else:
            raise exceptions.BadParametersException('Invalid "schedule" attribute')

    @property
    def data_reqs(self):
        return self._data_reqs

    @data_reqs.setter
    def data_reqs(self, value):
        if value is None:
            self._data_reqs = None
            return
        elif isinstance(value, DataRequirements):
            self._data_reqs = value
            return
        elif isinstance(value, dict):
            for key,reqs in value.items():
                if not isinstance(reqs, DataRequirements):
                    break
            else:
                self._data_reqs = value
                return
        raise exceptions.BadParametersException('Invalid data_reqs parameter')

    @property
    def allow_loops(self):
        return self._allow_loops

    @allow_loops.setter
    def allow_loops(self, value):
        self._allow_loops = bool(value)

    @property
    def m_in(self):
        return self._m_in

    def _initialize(self, owner):
        self.owner = owner
        self._register_metrics()

    def _inspect_input_params(self):
        self._p_in_routes = {}
        for k,o in self._p_in.items():
            if k in self._func_params:
                self._p_in_routes[k] = self._get_param_routes(obj=o, base=[], seen=set())

    def _inspect_output_params(self):
        self._p_out_routes = {}
        for k,o in self._p_out.items():
            if k in self._func_params and not k in self._p_in:
                self._p_out_routes[k] = self._get_param_routes(obj=o, base=[], seen=set())

    def _get_param_routes(self, obj, base, seen):
        param_routes = []
        my_id = id(obj)
        if my_id in seen:
            return param_routes
        else:
            seen.add(my_id)
        if isinstance(obj, Metric):
            if len(base) == 0:
                tmp_route = [('s',0)]
                param_routes.append(tmp_route)
            else:
                param_routes = base[:]
        elif isinstance(obj, list) or isinstance(obj,tuple):
            for i,item in enumerate(obj):
                tmp_route = base[:]
                tmp_route.append(('i',i))
                param_route = self._get_param_routes(obj=obj[i], base=tmp_route, seen=seen)
                if len(param_route)>0:
                    if isinstance(param_route[0],list):
                        for route in param_route:
                            param_routes.append(route)
                    else:
                        param_routes.append(param_route)
        elif isinstance(obj, dict):
            for k,v in obj.items():
                tmp_route = base[:]
                tmp_route.append(('k',k))
                param_route = self._get_param_routes(obj=obj[k], base=tmp_route, seen=seen)
                if len(param_route)>0:
                    if isinstance(param_route[0],list):
                        for route in param_route:
                            param_routes.append(route)
                    else:
                        param_routes.append(param_route)
        else:
            for att,value in inspect.getmembers(obj):
                if isinstance(value,list) or isinstance(value,tuple) or isinstance(value,dict) or isinstance(value,Metric):
                    tmp_route = base[:]
                    tmp_route.append(('a', att))
                    param_route = self._get_param_routes(obj=value, base=tmp_route, seen=seen)
                    if len(param_route)>0:
                        if isinstance(param_route[0],list):
                            for route in param_route:
                                param_routes.append(route)
                        else:
                            param_routes.append(param_route)
        return param_routes

    def _register_metrics(self):
        metrics = set()
        for k,r in self._p_in_routes.items():
            obj = self._p_in[k]
            for route in r:
                m = self._get_metric_by_route(obj=obj, route=route)
                if m:
                    metrics.add(m)
        self._m_in = metrics
        if self.schedule == None:
            self.schedule = OnUpdateSchedule(metrics = list(metrics))
        metrics = set()
        for k,r in self._p_out_routes.items():
            obj = self._p_out[k]
            for route in r:
                m = self._get_metric_by_route(obj=obj, route=route)
                if m:
                    metrics.add(m)
        self._m_out = metrics
        for m in self._m_out:
            if m in self._m_in and not self.allow_loops:
                logging.logger.warning('Posible loop in metric {} in transfer method {}. Loops are not allowed, output values for this metric will be discarded'.format(str(m.uri),str(f.__name__)))

    def _get_metric_by_route(self, obj, route):
        m = obj
        try:
            for p in route:
                if p[0] == 'a':
                    m = getattr(m,p[1],None)
                elif p[0] == 'k':
                    m = m.get(p[1],None)
                elif p[0] == 'i':
                    m = m[p[1]]
        except (AttributeError, ValueError, IndexError, TypeError) as e:
            logging.logger.debug('Exception _get_metric_by_route: {}'.format(str(e)))
            return None
        else:
            return m

    def get_data_requirements(self):
        reqs={}
        for m in self._m_in:
            reqs[m] = self.data_reqs[m.uri] if isinstance(self.data_reqs,dict) and m.uri in self.data_reqs else self.data_reqs
        return reqs

    def _get_execution_params(self, ts, metrics, data):
        exec_params={}
        arg_metrics_uris=[uri.get_global_uri(metric,owner=self.owner) for metric in metrics]
        updated = []
        others = []
        for arg in self._func_params:
            if arg == 'ts':
                exec_params[arg]=ts
            elif arg == 'updated':
                exec_params[arg]=[m for m in self._m_in if uri.get_global_uri(m,owner=self.owner) in arg_metrics_uris]
                for m in exec_params[arg]:
                    m.data = data[m]
            elif arg == 'others':
                exec_params[arg]=[m for m in self._m_in if uri.get_global_uri(m,owner=self.owner) not in arg_metrics_uris]
                for m in exec_params[arg]:
                    m.data = data[m]
            elif arg in self._p_in:
                #function input params, load metrics and add data attribute with the necessary data
                obj=copy.deepcopy(self._p_in[arg])
                if arg in self._p_in_routes:
                    for route in self._p_in_routes[arg]:
                        m = self._get_metric_by_route(obj=obj, route=route)
                        if m:
                            m.data = data[m]
                exec_params[arg]=obj
            elif arg in self._p_out:
                #funcion output params, load metrics and add data attribute with an empty Series() object
                obj=copy.deepcopy(self._p_out[arg])
                if arg in self._p_out_routes:
                    for route in self._p_out_routes[arg]:
                        m = self._get_metric_by_route(obj=obj, route=route)
                        if m:
                            m.data = pd.Series()
                exec_params[arg]=obj
        return exec_params

    async def _process_exec_result(self, params):
        samples = []
        for k,obj in params.items():
            if k in self._p_out:
                routes = None
                if k in self._p_in_routes:
                    routes = self._p_in_routes[k]
                elif k in self._p_out_routes:
                    routes = self._p_out_routes[k]
                if routes:
                    for route in routes:
                        m = self._get_metric_by_route(obj=obj, route=route)
                        if m and (self.allow_loops or not m in self._m_in):
                            for row in m.data.items():
                                samples.append(Sample(metric=m, ts=row[0], data=row[1]))
        if len(samples)>0:
            return {'samples':samples}

    def __call__(self, f):
        @wraps(f)
        async def decorated(ts, metrics, data):
            now=pd.Timestamp('now',tz='utc')
            exec_params=self._get_execution_params(ts=ts, metrics=metrics, data=data)
            self.last_exec=pd.Timestamp('now',tz='utc')
            if asyncio.iscoroutinefunction(f):
                await f(**exec_params)
            else:
                f(**exec_params)
            result = await self._process_exec_result(params=exec_params)
            return result
        self.f = decorated
        self._func_params = inspect.signature(f).parameters
        self._inspect_input_params()
        self._inspect_output_params()
        anon_transfer_methods.add_transfer_method(transfer_method=self, enabled=False)
        return f




import unittest
import uuid
import asyncio
import pandas as pd
from komlogd.api import crypto, session, exceptions

class ApiSessionTest(unittest.TestCase):

    def test_komlogsession_creation_failure_invalid_username(self):
        ''' creating a KomlogSession object should fail if user is invalid '''
        username = 'Invalid username'
        privkey=crypto.generate_rsa_key()
        with self.assertRaises(exceptions.BadParametersException) as cm:
            session.KomlogSession(username=username, privkey=privkey)
        self.assertEqual(cm.exception.msg, 'Invalid username {}'.format(str(username)))

    def test_komlogsession_creation_username_modification_not_allowed(self):
        ''' username cannot be modified on la KomlogSession object '''
        username = 'username'
        privkey=crypto.generate_rsa_key()
        s = session.KomlogSession(username=username, privkey=privkey)
        with self.assertRaises(exceptions.BadParametersException) as cm:
            s.username = 'new_username'
        self.assertEqual(cm.exception.msg, 'username modification not allowed')

    def test_komlogsession_creation_failure_invalid_privkey(self):
        ''' creating a KomlogSession object should fail if privkey is invalid '''
        username = 'username'
        privkey='privkey'
        with self.assertRaises(exceptions.BadParametersException) as cm:
            session.KomlogSession(username=username, privkey=privkey)
        self.assertEqual(cm.exception.msg, 'Invalid private key')

    def test_komlogsession_creation_privkey_modification_not_allowed(self):
        ''' username cannot be modified on la KomlogSession object '''
        username = 'username'
        privkey=crypto.generate_rsa_key()
        s = session.KomlogSession(username=username, privkey=privkey)
        with self.assertRaises(exceptions.BadParametersException) as cm:
            s.privkey = crypto.generate_rsa_key()
        self.assertEqual(cm.exception.msg, 'private key modification not allowed')




import sys
import unittest
from setuptools import setup, find_packages

version = '0.7.1'
if not sys.version_info[0] == 3 and not sys.version_info[1] >= 5:
    sys.exit("Sorry, Python 3.5 or greater required.")

print (find_packages())

def get_test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('lib/komlogd', pattern='test_*.py')
    return test_suite

setup(
    name='komlogd',
    version=version,
    description='Komlog agent',
    author='komlog Team',
    author_email='hello@komlog.io',
    url='https://github.com/komlog-io/komlogd',
    license='Apache Software License',
    install_requires=['pyyaml', 'setuptools', 'cryptography', 'aiohttp>=2.0', 'pandas>=0.18'],
    package_dir={ '': 'lib' },
    packages = find_packages('lib'),
    test_suite = 'setup.get_test_suite',
    scripts=[
       'bin/komlogd',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Build Tools',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ]
)



'''
logging methods

'''

import os
import logging
from komlogd.base import config, exceptions
from komlogd.base.settings import defaults,options
from logging.handlers import RotatingFileHandler

logger=None

def initialize_logger():
    global logger
    if logger:
        logger=None
    logger=logging.getLogger()
    log_config=config.config.get_logging_entries()
    if len(log_config)==0:
        log_level=defaults.LOG_LEVEL
        rotate_logs=defaults.LOG_ROTATION
        max_bytes=defaults.LOG_MAX_BYTES
        backup_count=defaults.LOG_BACKUP_COUNT
        log_format=defaults.LOG_FORMAT
        log_dir=os.path.join(config.config.root_dir,defaults.LOG_DIR)
        log_file=defaults.LOG_FILE
    else:
        log_level=log_config[0].get(options.LOG_LEVEL,defaults.LOG_LEVEL)
        rotate_logs=log_config[0].get(options.LOG_ROTATION,defaults.LOG_ROTATION)
        max_bytes=log_config[0].get(options.LOG_MAX_BYTES, defaults.LOG_MAX_BYTES)
        backup_count=log_config[0].get(options.LOG_BACKUP_COUNT, defaults.LOG_BACKUP_COUNT)
        log_format=log_config[0].get(options.LOG_FORMAT, defaults.LOG_FORMAT)
        log_dir=log_config[0].get(options.LOG_DIR, os.path.join(config.config.root_dir,defaults.LOG_DIR))
        if not os.path.isabs(log_dir):
            log_dir=os.path.join(config.config.root_dir,log_dir)
        log_file=log_config[0].get(options.LOG_FILE, defaults.LOG_FILE)
    try:
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        log_file=os.path.join(log_dir,os.path.split(log_file)[1])
        if rotate_logs is True:
            handler=RotatingFileHandler(log_file,'a',maxBytes=int(max_bytes),backupCount=int(backup_count))
        else:
            handler=logging.FileHandler(log_file)
        formatter=logging.Formatter(log_format)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(log_level)
        if not logger:
            raise exceptions.LoggerException()
        if len(log_config)>1:
            logger.info('Found more than one logger configuration, using the first one found.')
    except Exception as e:
        print(str(e))
        raise exceptions.LoggerException()




'''
File that implements configuration classes and methods

'''

import os
import yaml
import traceback
from komlogd.base import exceptions, logging
from komlogd.base.settings import defaults, templates, options

config=None

class Config():
    def __init__(self, filename):
        self.filename=filename
        self.root_dir = os.path.split(filename)[0]
        self.entries=None

    def _load_entries(self, filename=None, entryname=None):
        if filename is None:
            filename=self.filename
        try:
            entries=yaml.safe_load(open(filename,'r'))
        except Exception as e:
            raise exceptions.ConfigLoadException(str(e))
        if entries is None:
            return
        if entryname is None:
            for entry in entries:
                self.entries.append(entry)
        else:
            for entry in entries:
                if entryname in entry:
                    self.entries.append(entry)

    def get_entries(self, entryname):
        if self.entries is None:
            logging.logger.info('Loading configuration file entries')
            self.entries=[]
            self._load_entries()
        entries = []
        for entry in self.entries:
            if entryname in entry:
                entries.append(entry[entryname])
        return entries

    def get_logging_entries(self):
        ''' Before loading all configuration entries we initialize logging '''
        try:
            entries=yaml.safe_load(open(self.filename,'r'))
        except Exception as e:
            raise exceptions.ConfigLoadException(str(e))
        config = []
        if entries is None:
            return config
        for entry in entries:
            if options.ENTRY_LOG in entry:
                config.append(entry[options.ENTRY_LOG])
        return config


def load_application_config(filename=None):
    global config
    if not filename:
        filename=get_default_config_file()
        return load_application_config(filename)
    elif os.path.isfile(filename):
        config=Config(filename)
        if config:
            return True
    else:
        default_filename=get_default_config_file()
        if filename != default_filename:
            raise exceptions.ConfigLoadException('File not found: '+str(filename))
        elif _create_config_file(filename):
            return load_application_config(filename)
    raise exceptions.ConfigLoadException()

def get_default_config_file():
    home_dir = os.path.expanduser('~')
    filename = os.path.join(home_dir,defaults.APP_PATH,defaults.CONFIG_FILE)
    if not filename:
        raise exceptions.ConfigLoadException()
    return filename

def _create_config_file(filename):
    global config
    dirname,_ = os.path.split(filename)
    try:
        os.makedirs(dirname)
    except OSError as Err:
        if Err.errno == 17:
            pass
        else:
            print (str(Err))
            return False
    try:
        with open(filename,'wb') as f_out:
            f_out.write(bytes(templates.TEMPLATE_CONFIG_FILE,'UTF-8'))
            f_out.close()
        return True
    except Exception as e:
        ex_info=traceback.format_exc().splitlines()
        for line in ex_info:
            print(line)
        return False

